include( "Civ6Common" );

function OnInputHandler( pInputStruct:table )
	local debug = false;
	local uiMsg = pInputStruct:GetMessageType();
	if ( uiMsg == KeyEvents.KeyUp ) then 
		local uiKey:number = pInputStruct:GetKey();

		-- Check if numpad was used
		if ( uiKey ~= Keys.VK_NUMPAD1 and uiKey ~= Keys.VK_NUMPAD3 and
		     uiKey ~= Keys.VK_NUMPAD4 and uiKey ~= Keys.VK_NUMPAD6 and
		     uiKey ~= Keys.VK_NUMPAD7 and uiKey ~= Keys.VK_NUMPAD9 and
		     uiKey ~= Keys.VK_NUMPAD5 ) then
			return false;
		end
		
		-- Get the currently selected unit
		local unit:table = UI.GetHeadSelectedUnit();
		if ( unit == nil ) then return false; end

		local location          :table   = unit:GetLocation();
		local direction         :number  = 0;
		local new_x             :number  = 0;
		local new_y             :number  = 0;
		local nextPlot          :table;
		local destinationPlotID :number;
		local turnsList         :table;
		local obstacles         :table;
		local pathPlots         :table   = {};
		local allowMovement     :boolean = false;
		local isRangedAttack    :boolean = false;

		if debug then print("Numpad key pressed, trying to move unit"); end

		-- Get the direction to move to
		if ( uiKey == Keys.VK_NUMPAD1 ) then
			if debug then print("(Numpad 1) Moving south west"); end
			direction = DirectionTypes.DIRECTION_SOUTHWEST;
		
		elseif ( uiKey == Keys.VK_NUMPAD3 ) then
			if debug then print("(Numpad 3) Moving south east"); end
			direction = DirectionTypes.DIRECTION_SOUTHEAST;
		
		elseif ( uiKey == Keys.VK_NUMPAD4 ) then
			if debug then print("(Numpad 4) Moving west"); end
			direction = DirectionTypes.DIRECTION_WEST;

		elseif ( uiKey == Keys.VK_NUMPAD6 ) then
			if debug then print("(Numpad 6) Moving east"); end
			direction = DirectionTypes.DIRECTION_EAST;

		elseif ( uiKey == Keys.VK_NUMPAD7 ) then
			if debug then print("(Numpad 7) Moving north west"); end
			direction = DirectionTypes.DIRECTION_NORTHWEST;
	
		elseif ( uiKey == Keys.VK_NUMPAD9 ) then
			if debug then print("(Numpad 9) Moving north east"); end
			direction = DirectionTypes.DIRECTION_NORTHEAST;

		-- Center the map on unit
		elseif ( uiKey == Keys.VK_NUMPAD5 ) then
			if debug then print("(Numpad 5) Centering map on unit"); end
			UI.LookAtPlot(Map.GetPlot(location.x, location.y));
			return true;	-- break
		end


		-- Stop if unit cannot move on itself
		-- It's this low so that we can still center on the selected unit
		if ( GameInfo.Units[unit:GetUnitType()].IgnoreMoves ) then
			return false;
		end

		-- No movement left
		if ( unit:GetMovesRemaining() == 0 ) then 
			return false;
		end

		-- Get the next plot in the selected direction
		nextPlot = Map.GetAdjacentPlot(location.x, location.y, direction);
		if ( nextPlot == nil ) then
			if debug then print("No adjacent plot found"); end
			return false;
		end

		new_x = nextPlot:GetX();
		new_y = nextPlot:GetY();

		if debug then print("New X: ", new_x); end
		if debug then print("New Y: ", new_y); end

		-- Get the path the unit would need to take to move to the next plot
		pathPlots, turnsList, obstacles = UnitManager.GetMoveToPath(unit, Map.GetPlotIndex(new_x, new_y));


		-- Do not move unit if more than one plot has to be crossed (i.e. due to obstacles like cliffs or a Zone of Control)
		if debug then print("Number of plots in the path: ", table.count(pathPlots)); end
		if ( table.count(pathPlots) > 2 ) then
			if debug then print("Path would need more than 1 movement action"); end
			return false;
		end

		-- Do not move unit if more than one turn would be needed to move to the plot (i.e. don't queue the movement)
		-- The nil check handles ranged units when trying to move onto an attackable plot
		if debug then print("Number of turns for the last move: ", turnsList[2]); end
		-- if ( turnsList[2] == nil or turnsList[2] > 1 ) then return false; end;
		if ( turnsList[2] ~= nil and turnsList[2] > 1 ) then
			if debug then print("Movement would take more than 1 turn"); end
			return false;
		end

		-- The final destination plot for the movement, determined by the path
		destinationPlotID = pathPlots[table.count(pathPlots)];

		-- No plot in the in the path found
		if ( destinationPlotID == nil ) then
			-- Is a ranged attack possible for that tile?
			if debug then print("Checking for ranged attack"); end
			local tParameters:table = {};
			tParameters[UnitOperationTypes.PARAM_X] = new_x;
			tParameters[UnitOperationTypes.PARAM_Y] = new_y;
			tParameters[UnitOperationTypes.PARAM_MODIFIERS] = UnitOperationMoveModifiers.NONE;
			local bCanStart = UnitManager.CanStartOperation( unit, UnitOperationTypes.RANGE_ATTACK, nil, tParameters, true);

			if debug then print("Can initiate a ranged attack: ", bCanStart); end

			-- Valid ranged target
			if ( bCanStart ) then
				allowMovement  = true;
				isRangedAttack = true;
			-- Cannot initiate a ranged attack, break as it is no valid plot to move to
			else
				if debug then print("No final destination plot in the path found"); end
				-- return false;	-- Breaking here prevents the War popup
			end
		end


		-- Unit has just moved into a Zone of Control, and is only allowed to attack the target for the rest of this turn
		-- Do not allow any other movements, as they would be queued for the next turn
		if debug then print("unit:HasMovedIntoZOC(): ", unit:HasMovedIntoZOC()); end
		if ( unit:HasMovedIntoZOC() ) then
			local kAttackPlots :table = UnitManager.GetReachableTargets(unit);

			if debug then print("destinationPlotID: ", destinationPlotID); end

			if ( not allowMovement and kAttackPlots ~= nil and table.count( kAttackPlots ) > 0 ) then
				for _, plot in ipairs( kAttackPlots ) do
					if debug then print("attack plot:       ", plot); end
					
					-- The destination plot is the same as an attackable plot, so allow the movement
					if ( plot == destinationPlotID ) then
						if debug then print("Destination plot is an attackable plot, allow movement"); end
						allowMovement = true;
						break;
					end
				end
			end

			-- Sometimes it bugs out, and the plot to attack on is not marked as reachable
			-- Let's check if the destinationPlotID is an enemy target
			if ( not allowMovement ) then
				local bIsEnemyAtEnd :boolean = false;
				local pLocalPlayer	:number  = Game.GetLocalPlayer();
				local plotCity               = Cities.GetCityInPlot(new_x, new_y);
				local plotDistrict  :number  = -1;
				local otherPlayerID :number  = -1;
				local bIsAtWarWith  :boolean = false;
				
				-- Check the plot for an enemy city
				if ( plotCity ~= nil and plotCity:GetOwner() ~= pLocalPlayer ) then
					otherPlayerID = plotCity:GetOwner();
					bIsAtWarWith  = Players[pLocalPlayer]:GetDiplomacy():IsAtWarWith(otherPlayerID);
					
					if ( bIsAtWarWith ) then
						if debug then print("Destination plot is an enemy city, allow movement"); end
						allowMovement = true;
					end
				end

				-- Check for Encampments
				-- print("nextPlot:GetDistrictType(): ", nextPlot:GetDistrictType());
				-- print("District NameType:          ", GameInfo.Districts[nextPlot:GetDistrictType()]);
				-- print("Plot Owner ID:              ", nextPlot:GetOwner());
				plotDistrict = nextPlot:GetDistrictType();
				if ( plotDistrict > -1 and GameInfo.Districts[plotDistrict].DistrictType == "DISTRICT_ENCAMPMENT" ) then
					otherPlayerID = nextPlot:GetOwner();
					bIsAtWarWith  = Players[pLocalPlayer]:GetDiplomacy():IsAtWarWith(otherPlayerID);

					if ( bIsAtWarWith ) then
						if debug then print("Destination plot is an enemy encampment, allow movement"); end
						allowMovement = true;
					end
				end

				
				-- Check for units
				--[[
				if ( endPlot ~= nil ) then
					local unitList = Units.GetUnitsInPlotLayerID( endPlot:GetX(), endPlot:GetY(), MapLayers.ANY );
					for i, pUnit in ipairs(unitList) do
						if ( pLocalPlayer ~= pUnit:GetOwner() and pPlayerVis ~= nil and pPlayerVis:IsVisible(endPlot:GetX(), endPlot:GetY()) and pPlayerVis:IsUnitVisible(pUnit) ) then
							bIsEnemyAtEnd = true;
							allowMovement = true;
						end
					end
				end
				--]]
			end


			-- Don't move if not attacking
			if ( not allowMovement ) then
				if debug then print("No possible movement actions found"); end
				return false;
			-- Check if the action is allowed
			elseif (not isRangedAttack ) then
				local tParameters:table = {};
				tParameters[UnitOperationTypes.PARAM_X] = new_x;
				tParameters[UnitOperationTypes.PARAM_Y] = new_y;
				tParameters[UnitOperationTypes.PARAM_MODIFIERS] = UnitOperationMoveModifiers.ATTACK;
				local bCanStart, tResults = UnitManager.CanStartOperation( unit, UnitOperationTypes.MOVE_TO, nil, tParameters, true);
				if debug then print("UnitManager.CanStartOperation: "); end
				if debug then print("bCanStart: ", bCanStart); end

				-- Break if not allowed to arrack
				if ( not bCanStart ) then
					return false;
				end
			end
		end

		-- Move the unit
		-- This will trigger a Confirm War popup if necessary (function is in Civ6Common.lua)
		if debug then print("Trying to move the unit"); end
		MoveUnitToPlot(unit, new_x, new_y);
	end
end


-- Initialize
function Initialize()
	ContextPtr:SetInputHandler(OnInputHandler, true);	-- Not sure why true, but found it this way in a lua file
end


Initialize();